@@ -1,7 +1,7 @@
 # --------------------------------------------------------------- Imports ---------------------------------------------------------------- #
 
 # System
-from typing import List, Dict, Optional, Tuple
+from typing import List, Dict, Optional, Tuple, Callable
 import time, json
 
 # Pip
@@ -11,6 +11,9 @@ from kcu import strings
 
 from bs4 import BeautifulSoup as bs
 
+# Local
+from .visibility import Visibility
+
 # ---------------------------------------------------------------------------------------------------------------------------------------- #
 
 
@@ -48,13 +51,19 @@ class Youtube:
         port: Optional[int] = None,
         screen_size: Optional[Tuple[int, int]] = None,
         full_screen: bool = False,
-        disable_images: bool = False
+        disable_images: bool = False,
+        login_prompt_callback: Optional[Callable[[str], None]] = None
     ):
         self.browser = Firefox(cookies_folder_path, extensions_folder_path, host=host, port=port, screen_size=screen_size, full_screen=full_screen, disable_images=disable_images)
         self.channel_id = None
 
         try:
-            if self.browser.login_via_cookies(YT_URL, LOGIN_INFO_COOKIE_NAME) or self.login(email=email, password=password):
+            self.__internal_channel_id = cookies_folder_path.strip('/').split('/')[-1]
+        except:
+            self.__internal_channel_id = cookies_folder_path
+
+        try:
+            if self.browser.login_via_cookies(YT_URL, LOGIN_INFO_COOKIE_NAME) or self.login(email=email, password=password, login_prompt_callback=login_prompt_callback):
                 time.sleep(0.5)
                 self.browser.get(YT_URL)
                 time.sleep(0.5)
@@ -77,7 +86,12 @@ class Youtube:
 
     # -------------------------------------------------------- Public methods -------------------------------------------------------- #
 
-    def login(self, email: Optional[str], password: Optional[str]) -> bool:
+    def login(
+        self,
+        email: Optional[str],
+        password: Optional[str],
+        login_prompt_callback: Optional[Callable[[str], None]] = None
+    ) -> bool:
         org_url = self.browser.driver.current_url
         self.browser.get(YT_URL)
         time.sleep(0.5)
@@ -89,7 +103,7 @@ class Youtube:
         
         if not logged_in:
             print('Could not log you in automatically.')
-            logged_in = self.__login_manual()
+            logged_in = self.__login_manual(login_prompt_callback=login_prompt_callback)
 
         if logged_in:
             time.sleep(0.5)
@@ -109,20 +123,25 @@ class Youtube:
         title: str,
         description: str,
         tags: List[str],
+        made_for_kids: bool = False,
+        visibility: Visibility = Visibility.PUBLIC,
+        thumbnail_image_path: Optional[str] = None,
         _timeout: Optional[int] = 60*3, # 3 min
-        extra_sleep_after_upload: Optional[int] = None
+        extra_sleep_after_upload: Optional[int] = None,
+        extra_sleep_before_publish: Optional[int] = None
     ) -> (bool, Optional[str]):
+
         if _timeout is not None:
             try:
                 with timeout.timeout(_timeout):
-                    return self.__upload(video_path, title, description, tags, extra_sleep_after_upload=extra_sleep_after_upload)
+                    return self.__upload(video_path, title, description, tags, made_for_kids=made_for_kids, visibility=visibility, thumbnail_image_path=thumbnail_image_path, extra_sleep_after_upload=extra_sleep_after_upload, extra_sleep_before_publish=extra_sleep_before_publish)
             except Exception as e:
                 print('Upload', e)
                 # self.browser.get(YT_URL)
 
                 return False, None
         else:
-            return self.__upload(video_path, title, description, tags, extra_sleep_after_upload=extra_sleep_after_upload)
+            return self.__upload(video_path, title, description, tags, made_for_kids=made_for_kids, visibility=visibility, thumbnail_image_path=thumbnail_image_path, extra_sleep_after_upload=extra_sleep_after_upload, extra_sleep_before_publish=extra_sleep_before_publish)
 
     def get_current_channel_id(self) -> Optional[str]:
         self.browser.get(YT_URL)
@@ -246,7 +265,11 @@ class Youtube:
         title: str,
         description: str,
         tags: List[str],
-        extra_sleep_after_upload: Optional[int] = None
+        made_for_kids: bool = False,
+        visibility: Visibility = Visibility.PUBLIC,
+        thumbnail_image_path: Optional[str] = None,
+        extra_sleep_after_upload: Optional[int] = None,
+        extra_sleep_before_publish: Optional[int] = None
     ) -> (bool, Optional[str]):
         self.browser.get(YT_URL)
         time.sleep(1.5)
@@ -259,6 +282,9 @@ class Youtube:
             self.browser.find(By.XPATH, "//input[@type='file']").send_keys(video_path)
             print('Upload: uploaded video')
 
+            if extra_sleep_after_upload is not None and extra_sleep_after_upload > 0:
+                time.sleep(extra_sleep_after_upload)
+
             title_field = self.browser.find_by('div', id_='textbox')
             time.sleep(0.5)
             title_field.send_keys(Keys.BACK_SPACE)
@@ -283,9 +309,21 @@ class Youtube:
             tags_field.send_keys(','.join(tags) + ',')
             print("Upload: added tags")
 
-            kids_section = self.browser.find(By.NAME, "NOT_MADE_FOR_KIDS")
+            if made_for_kids:
+                kids_selection_name = "MADE_FOR_KIDS"
+            else:
+                kids_selection_name = "NOT_MADE_FOR_KIDS"
+
+            kids_section = self.browser.find(By.NAME, kids_selection_name)
             self.browser.find(By.ID, "radioLabel", kids_section).click()
-            print("Upload: did set NOT_MADE_FOR_KIDS")
+            print('Upload: did set', kids_selection_name)
+
+            if thumbnail_image_path is not None:
+                try:
+                    self.browser.find_by('input', id_='file-loader')
+                    time.sleep(0.5)
+                except Exception as e:
+                    print('Upload: Thumbnail error: ', e)
             
             self.browser.find(By.ID, 'next-button').click()
             print('Upload: clicked first next')
@@ -293,9 +331,9 @@ class Youtube:
             self.browser.find(By.ID, 'next-button').click()
             print('Upload: clicked second next')
 
-            public_main_button = self.browser.find(By.NAME, "PUBLIC")
-            self.browser.find(By.ID, 'radioLabel', public_main_button).click()
-            print('Upload: set to public')
+            visibility_main_button = self.browser.find(By.NAME, visibility.name)
+            self.browser.find(By.ID, 'radioLabel', visibility_main_button).click()
+            print('Upload: set to', visibility.name)
 
             try:
                 video_url_container = self.browser.find(By.XPATH, "//span[@class='video-url-fadeable style-scope ytcp-video-info']", timeout=2.5)
@@ -307,33 +345,49 @@ class Youtube:
 
             i=0
 
-            if extra_sleep_after_upload is not None and extra_sleep_after_upload > 0:
-                time.sleep(extra_sleep_after_upload)
+            if extra_sleep_before_publish is not None and extra_sleep_before_publish > 0:
+                time.sleep(extra_sleep_before_publish)
 
             while True:
                 try:
-                    done_button = self.browser.find(By.ID, 'done-button')
+                    progress_label = self.browser.find_by('span', { 'class': 'progress-label style-scope ytcp-video-upload-progress' }, timeout=0.1) or self.browser.find_by('span', { 'class': 'progress-label-old style-scope ytcp-video-upload-progress' }, timeout=0.1)
+                    progress_label_text = progress_label.text.lower()
+                    
+                    if 'finished' in progress_label_text or 'process' in progress_label_text:
+                        done_button = self.browser.find(By.ID, 'done-button')
 
-                    if done_button.get_attribute('aria-disabled') == 'false':
-                        done_button.click()
+                        if done_button.get_attribute('aria-disabled') == 'false':
+                            done_button.click()
 
-                        print('Upload: published')
+                            print('Upload: published')
 
-                        time.sleep(3)
-                        self.browser.get(YT_URL)
+                            time.sleep(3)
+                            self.browser.get(YT_URL)
 
-                        return True, video_id
+                            return True, video_id
                 except Exception as e:
                     print(e)
-
                     i += 1
 
-                    if i >= 10:
+                    if i >= 20:
+                        done_button = self.browser.find(By.ID, 'done-button')
+
+                        if done_button.get_attribute('aria-disabled') == 'false':
+                            done_button.click()
+
+                            print('Upload: published')
+
+                            time.sleep(3)
+                            self.browser.get(YT_URL)
+
+                            return True, video_id
+
                         raise
 
                 time.sleep(1)
         except Exception as e:
             print(e)
+                
             self.browser.get(YT_URL)
 
             return False, None
@@ -343,6 +397,10 @@ class Youtube:
         self.load_video(video_id)
         time.sleep(1)
         self.browser.scroll(150)
+        time.sleep(1)
+        self.browser.scroll(100)
+        time.sleep(1)
+        self.browser.scroll(100)
 
         try:
             header = self.browser.find_by('div', id_='masthead-container', class_='style-scope ytd-app')
@@ -470,11 +528,14 @@ class Youtube:
 
         return self.is_logged_in
     
-    def __login_manual(self) -> bool:
+    def __login_manual(self, login_prompt_callback: Optional[Callable[[str], None]] = None) -> bool:
         self.browser.get(YT_LOGIN_URL)
         time.sleep(0.5)
 
         try:
+            if login_prompt_callback is not None:
+                login_prompt_callback(self.__internal_channel_id + ' needs manual input to log in.')
+
             input('Log in then press return')
         except Exception as e:
             print('__login_manual', e)
